import hashlib
import random
from .utils import LookupTable


class Provider:
    def meme(self, name: str) -> str:
        pass


class ConsistentProvider(Provider):
    """
    This provider will assign memes based on the name of the identifier and nothing else.
    """

    def __init__(self, available_memes: list):
        self.__lookup_table = LookupTable(self.__hash, available_memes)
        self.__available_memes = available_memes

    def __hash(self, name: str) -> int:
        return int(hashlib.sha256(name.encode("utf-8")).hexdigest(), 16) % len(
            self.__available_memes
        )

    def meme(self, name: str) -> str:
        return self.__lookup_table[name]


class RandomConsistentProvider(Provider):
    """
    This provider will assign a random meme based on the name of the identifier. It is consistent because the same name in the same session will always result in the same meme.
    """

    def __init__(self, available_memes: list):
        self.__lookup_table = LookupTable(self.__hash, available_memes)
        self.__available_memes = available_memes
        self.__name_to_hash_number = dict()

    def __hash(self, name: str) -> int:
        if name in self.__name_to_hash_number:
            return self.__name_to_hash_number[name]
        else:
            h = random.randrange(len(self.__available_memes))
            self.__name_to_hash_number[name] = h
            return h

    def meme(self, name: str) -> str:
        return self.__lookup_table[name]


class SequentialProvider(Provider):
    """
    This provider will assign memes based on the order it encounters names in the source code. This provider ensures that all names in the meme dictionary will be used if there are more unique names than entries in the meme dictionary.
    """

    def __init__(self, available_memes: list):
        self.__lookup_table = LookupTable(self.__hash, available_memes)
        self.__available_memes = available_memes
        self.__name_to_hash_number = dict()

    def __hash(self, name: str) -> int:
        if name in self.__name_to_hash_number:
            return self.__name_to_hash_number[name]
        else:
            h = len(self.__name_to_hash_number)
            self.__name_to_hash_number[name] = h
            return h

    def meme(self, name: str) -> str:
        return self.__lookup_table[name]


class RandomSequentialProvider(Provider):
    """
    This provider will assign memes based on the randomised order it encounters names in the source code. This provider ensures that all names in the meme dictionary will be used if there are more unique names than entries in the meme dictionary.
    """

    def __init__(self, available_memes: list):
        self.__lookup_table = LookupTable(self.__hash, available_memes)
        self.__available_memes = available_memes
        self.__name_to_hash_number = dict()

    def __hash(self, name: str) -> int:
        if name in self.__name_to_hash_number:
            return self.__name_to_hash_number[name]
        else:
            unused_numbers = set(range(len(self.__available_memes))) - set(
                self.__name_to_hash_number.values()
            )
            if len(unused_numbers) != 0:
                h = random.sample(unused_numbers, 1)[0]
                self.__name_to_hash_number[name] = h
                return h
            else:
                # no more unused numbers, just random
                h = random.randrange(len(self.__available_memes))
                self.__name_to_hash_number[name] = h
                return h

    def meme(self, name: str) -> str:
        return self.__lookup_table[name]



class LookupTable:
    class Node:
        def __init__(self, name):
            self.name = name
            self.next = None

    def __init__(self, hash_function: callable, memes: list):
        self.__name_table = dict()
        self.__meme_table = dict(enumerate(memes))
        self.__hash = hash_function

    def __getitem__(self, name: str):
        if type(name) is not str:
            raise TypeError(f"Expected str for name, got {type(name)}")
        h = self.__hash(name)
        if h in self.__name_table:
            node = self.__name_table[h]
            depth = 0
            while node.next is not None and node.name != name:
                node = node.next
                depth += 1
            if node.next is None and node.name != name:
                return self.__init_item(name) + "_copy" * (depth + 1)
            if h not in self.__meme_table:
                raise IndexError(f"Index {h} does not exist in Meme Lookup Table")
            return self.__meme_table[h] + "_copy" * depth
        else:
            return self.__init_item(name)

    def __init_item(self, name: str):
        h = self.__hash(name)
        if h in self.__name_table:
            node = self.__name_table[h]
            depth = 0
            while node.next is not None and node.name != name:
                node = node.next
                depth += 1
            if node.name == name:
                return
            elif node.next is None:
                node.next = self.__class__.Node(name)
            else:
                raise LookupError("Illegal state")
        else:
            self.__name_table[h] = self.__class__.Node(name)
        return self.__meme_table[h]



import argparse
import tokenize
import re
import typing
import pkgutil
import io
from . import gatekeepers
from . import providers


class Obfuscator:
    def obfuscate(
        self, source: typing.IO, output: typing.IO, provider: providers.Provider
    ):
        output_tokens = []
        gatekeeper = gatekeepers.SafeGatekeeper()
        offset = 0
        encoding = None
        for token in tokenize.tokenize(source.readline):
            if token.type == tokenize.ENCODING:
                encoding = token.string
            """
            if token.type == tokenize.NAME and token.string[0].lower() == "f":
                inner_source = io.StringIO()
                self.obfuscate()
            """
            if gatekeeper.read(token):
                meme = provider.meme(token.string)
                sub_offset = len(meme) - len(token.string)
                output_tokens.append(
                    (
                        tokenize.NAME,
                        meme,
                        (token.start[0], token.start[1] + offset),
                        (token.end[0], token.end[1] + offset + sub_offset),
                        token.line,
                    )
                )
                offset += sub_offset
            else:
                output_tokens.append(
                    (
                        token.type,
                        token.string,
                        (token.start[0], token.start[1] + offset),
                        (token.end[0], token.end[1] + offset),
                        token.line,
                    )
                )
            if token.type == tokenize.NEWLINE:
                offset = 0
        output.write(tokenize.untokenize(output_tokens).decode(encoding))


def main():
    parser = argparse.ArgumentParser(
        description="Obfuscates variables in collaboration with MIT."
    )

    parser.add_argument(
        "input_file", type=argparse.FileType("rb"), help="the Python file to obfuscate"
    )

    parser.add_argument(
        "-o",
        "--output-file",
        type=argparse.FileType("w"),
        required=False,
        help="the destination file",
    )

    parser.add_argument(
        "--random",
        action="store_true",
        required=False,
        help="makes the obfuscation process non deterministic",
    )

    parser.add_argument(
        "--sequential",
        "--seq",
        action="store_true",
        required=False,
        help="ensures that all memes in the dictionary are used before recycling names",
    )

    parser.add_argument(
        "--memes",
        required=False,
        help="custom dictionary file for retrieving replacement names in obfuscation",
    )

    args = vars(parser.parse_args())

    input_file = args.get("input_file")
    output_file = args.get("o") or open(
        re.sub(r"\.py$", ".sutd.py", input_file.name), "w"
    )

    try:
        if args.get("memes") is None:
            raise FileNotFoundError
        builtin_dictionary_file = pkgutil.get_data(
            "sutdobfs", "memes/" + args.get("memes")
        )
        dictionary_file = io.StringIO(builtin_dictionary_file.decode("utf-8"))
    except FileNotFoundError:
        try:
            if args.get("memes") is None:
                raise FileNotFoundError
            user_supplied_dictionary_file = open(args.get("memes"))
            dictionary_file = user_supplied_dictionary_file
        except FileNotFoundError:
            # use default memes.txt
            dictionary_file = io.StringIO(
                pkgutil.get_data("sutdobfs", "memes/memes.txt").decode("utf-8")
            )

    memes = [
        line.strip()
        for line in dictionary_file.readlines()
        if str.isidentifier(line.strip())
    ]

    if args.get("random") and args.get("sequential"):
        provider = providers.RandomSequentialProvider(memes)
    elif args.get("random"):
        provider = providers.RandomConsistentProvider(memes)
    elif args.get("sequential"):
        provider = providers.SequentialProvider(memes)
    else:
        provider = providers.ConsistentProvider(memes)

    obfs = Obfuscator()

    obfs.obfuscate(input_file, output_file, provider)

    input_file.close()
    output_file.close()

    return 0


if __name__ == "__main__":
    code = main()
    exit(code)



import keyword
import tokenize


class SafeGatekeeper:
    """
  This class will only meme non-member identifiers in the Local and Enclosed scopes
  """

    def __init__(self):
        self.__scope_state = "global"
        self.__member_state = "nonmember"
        self.globals = []
        self.indent_count = 0

    def read(self, token):
        # first handle scope state transitions
        if self.__scope_state == "global":
            if token.type == tokenize.NAME and token.string == "def":
                self.__scope_state = "def"
        elif self.__scope_state == "def":
            if token.type == tokenize.OP and token.string == ":":
                self.__scope_state = ":"
        elif self.__scope_state == ":":
            if token.type == tokenize.NEWLINE:
                self.__scope_state = "local"
            else:
                self.__scope_state = "def"
        elif self.__scope_state == "local":
            if token.type == tokenize.INDENT:
                self.indent_count += 1
            elif token.type == tokenize.DEDENT:
                self.indent_count -= 1
                if self.indent_count <= 0:
                    self.__scope_state = "global"
                    self.indent_count = 0
        # next handle member/nonmember state transitions
        if self.__member_state == "nonmember":
            if token.type == tokenize.OP and token.string == ".":
                self.__member_state = "member"
        elif self.__member_state == "member":
            if token.type == tokenize.NAME:
                self.__member_state = "member"
            elif token.type == tokenize.OP and token.string == ".":
                self.__member_state = "member"
            else:
                self.__member_state = "nonmember"
        # register globals
        if self.__scope_state != "local":
            if token.type == tokenize.NAME:
                if str.isidentifier(token.string) and not keyword.iskeyword(
                    token.string
                ):
                    self.globals.append(token.string)
        # then handle output
        if self.__scope_state != "local" or self.__member_state != "nonmember":
            return False
        if (
            token.type == tokenize.NAME
            and str.isidentifier(token.string)
            and not keyword.iskeyword(token.string)
            and token.string not in __builtins__
            and token.string not in self.globals
        ):
            return True
        else:
            return False



from setuptools import find_packages, setup

with open("README.md", "rb") as f:
    long_descr = f.read().decode("utf-8")


setup(
    name="sutdobfs",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    entry_points={"console_scripts": ["sutdobfs = sutdobfs.__main__:main"]},
    version="1.0.3",
    description="Establish your variable names in collaboration with MIT",
    long_description=long_descr,
    long_description_content_type="text/markdown",
    author="Chester Koh",
    author_email="chester8991@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    url="https://github.com/OpenSUTD/sutdobfs",
)
